;;;
;;; scmail.bayesian-filter - Bayesian spam filter experiment
;;;
;;;  Copyright(C) 2003 by Shiro Kawai (shiro@acm.org)
;;;
;;;  Permission to use, copy, modify, distribute this software and
;;;  accompanying documentation for any purpose is hereby granted,
;;;  provided that existing copyright notices are retained in all
;;;  copies and that this notice is included verbatim in all
;;;  distributions.
;;;  This software is provided as is, without express or implied
;;;  warranty.  In no circumstances the author(s) shall be liable
;;;  for any damages arising out of the use of this software.
;;;
;;;  $Id: bayesian-filter.scm,v 1.2 2003/06/17 07:56:50 shirok Exp $

(define-module scmail.bayesian-filter
  (use srfi-2)
  (use srfi-13)
  (use rfc.822)
  (use rfc.base64)
  (use rfc.quoted-printable)
  (use gauche.charconv)
  (use gauche.parameter)
  (use gauche.sequence)
  (use file.util)
  (use binary.pack)
  (use util.list)
  (use dbm)
  (use dbm.gdbm) ;; should be customizable
  (export prob-tables spam-table nonspam-table with-token-table
          save-prob-tables load-prob-tables
          collect-words spamness-of-word spamness-of-mail mail-is-spam?
          simple-stat test-spamness-of-mail test-spamness-in-folder
          convert-database)
  )
(select-module scmail.bayesian-filter)

;;==========================================================
;; Probability table
;;

(define-constant *dbm-class* <gdbm>) ;; should be customizable

(define-constant *message-count-key* " %message-count")
(define-constant *total-token-count-key* " %total-token-count")

(define-constant *languages* '(#t jp))
(define-constant *num-values* (* (length *languages*) 2))
(define-constant *value-packer* (make-string *num-values* #\V))

(define-class <token-table> ()
  ((db     :init-keyword :db)
   (rwmode :init-keyword :rwmode)
   (cache  :initform (make-hash-table 'string=?))
   (message-count :init-keyword :message-count)
   (total-token-count :init-keyword :total-token-count)
   ))

(define (pack-value v)
  (pack *value-packer* (vector->list v) :to-string? #t))
(define (unpack-value s)
  (list->vector (unpack *value-packer* :from-string s)))

(define (open-token-table file rwmode)
  (define (get-values db key)
    (cond ((dbm-get db key #f) => unpack-value)
          (else (make-vector *num-values* 0))))
  (let* ((db     (dbm-open *dbm-class* :path file :rw-mode rwmode))
         (mcount (get-values db *message-count-key*))
         (tcount (get-values db *total-token-count-key*)))
    (make <token-table> :db db :rwmode rwmode
          :message-count mcount  :total-token-count tcount)))

(define (close-token-table tab)
  (define z 0)
  (when (memv (ref tab 'rwmode) '(:write :create))
    (hash-table-for-each
     (ref tab 'cache)
     (lambda (k v)
       (let* ((s (dbm-get (ref tab 'db) k #f))
              (newv (if s
                        (map-to <vector> + v (unpack-value s))
                        v)))
         (inc! z)
         (dbm-put! (ref tab 'db) k (pack-value newv)))))
    (for-each (lambda (key slot)
                (dbm-put! (ref tab 'db) key (pack-value (ref tab slot))))
              (list *message-count-key* *total-token-count-key*)
              '(message-count total-token-count))
    )
  (dbm-close (ref tab 'db)))

(define token-table (make-parameter #f))

(define (with-token-table file rwmode thunk)
  ;; NB: this should really be an unwind-protect.
  ;; Waiting for its implementation...
  (dynamic-wind
   (lambda () (token-table (open-token-table file rwmode)))
   thunk
   (lambda () (close-token-table (token-table)))))

;; Open token table for reading.
;; We keep the name for backward compatibility.
(define (load-prob-tables file)
  (let1 realfile (cond ((string-suffix? ".dbm" file) file)
                       ((file-exists? #`",|file|.dbm") #`",|file|.dbm")
                       ((string=? (string-complete->incomplete
                                   (with-input-from-file file
                                     (cut read-block 10)))
                                  #*";;-*-Schem")
                        (errorf "Database file ~s seems to have an old format.  Run scbayes --update-db" file))
                       (else file))
    (token-table (open-token-table realfile :read))))

(define (add-token type token . maybe-count)
  (let ((tab (token-table))
        (cnt (get-optional maybe-count 1)))
    (or (and-let* ((v (hash-table-get (ref tab 'cache) token #f)))
          (inc! (ref v type) cnt)
          (when (zero? (ref v type))
            (inc! (ref (ref tab 'total-token-count) type))))
        (let1 v (make-vector *num-values* 0)
          (inc! (ref v type) cnt)
          (inc! (ref (ref tab 'total-token-count) type))
          (hash-table-put! (ref tab 'cache) token v)))
    ))

(define (token-count type token)
  (let* ((tab (token-table))
         (v   (or (hash-table-get (ref tab 'cache) token #f)
                  (and-let* ((s (dbm-get (ref tab 'db) token #f))
                             (v (unpack-value s)))
                    (hash-table-put! (ref tab 'cache) token v)
                    v))))
    (if v (ref v type) 0)))

(define (message-count type)
  (ref (ref (token-table) 'message-count) type))

(set! (setter message-count)
      (lambda (type val)
        (set! (ref (ref (token-table) 'message-count) type) val)))

;; get appropriate table type index
(define (spam-table lang)
  (+ (* (or (find-index (cut eqv? <> lang) *languages*) 0) 2) 1))

(define (nonspam-table lang)
  (* (or (find-index (cut eqv? <> lang) *languages*) 0) 2))

;; Convert old database to new database
(define (convert-database old-path new-path)
  (define current-table #f)
  (with-input-from-file old-path
    (lambda ()
      (with-token-table
       new-path :create
       (lambda ()
         (port-for-each
          (lambda (expr)
            (if (keyword? (car expr))
                (let1 tab ((if (get-keyword :spam? expr)
                               spam-table
                               nonspam-table)
                           (get-keyword :lang expr))
                  (set! (message-count tab)
                        (get-keyword :message-count expr))
                  (set! current-table tab))
                (add-token current-table (car expr) (cdr expr))))
          read)))))
  )

;;==========================================================
;; Collecting statistics
;;

(define word-probability-limit 0.001)

;; Count tokens from emails in the folder.
;;   (collect-words "~/Mail/spam"  spam-table)
;;   (collect-words "~/Mail/inbox" nonspam-table)
(define (collect-words directory table-getter)
  (for-each (lambda (email)
              (receive (tokens lang) (tokenize-email email)
                (unless (null? tokens)
                  (let1 tab (table-getter lang)
                    (for-each (cut add-token tab <>) tokens)
                    (inc! (message-count tab))))))
            (directory-list (expand-path directory)
                            :children? #t :add-path? #t
                            :filter #/^\d+$/)))

;; Calculate a word's spam probability.
;;  Cf. Paul Graham, "A Plan for Spam" and "Better Bayesian Filtering"
(define (spamness-of-word word lang)
  (let* ((gtab (nonspam-table lang))
         (btab (spam-table lang))
         (g (token-count gtab word))
         (b (token-count btab word))
         (min-p word-probability-limit)
         (min-p2 (* min-p 2)))
    (cond ((< (+ (* g 2) b) 5)
           0.4) ; default 'spamness' for unfamiliar word
          ((zero? g)
           (if (> b 10) (- 1 min-p) (- 1 min-p2))) ;see "Better" paper
          ((zero? b)
           (if (> g 10) min-p min-p2)) ;see "Better" paper
          (else
           (clamp (/ (clamp (/ b (message-count btab)) #f 1.0)
                     (+ (clamp (/ (* g 2) (message-count gtab)) #f 1.0)
                        (clamp (/ b (message-count btab)) #f 1.0)))
                  min-p2
                  (- 1 min-p2)))
          )))

;; Given list of probabilities, calculates combined probability
;;  using Bayesian probability.
(define (spam-probability probs)
  (let1 prod (apply * probs)
    (/ prod (+ prod (apply * (map (cut - 1 <>) probs))))))

;; Take NUM most-interesting words from the list of tokens.
(define (most-interesting-words lang tokens num)
  (let ((tab (make-hash-table 'string=?))
        (extremeness (lambda (prob) (abs (- 0.5 prob)))))
    (for-each (lambda (tok)
                (unless (hash-table-get tab tok #f)
                  (hash-table-put! tab tok (spamness-of-word tok lang))))
              tokens)
    (let ((wlist (sort (hash-table-map tab cons)
                       (lambda (a b) (> (extremeness (cdr a))
                                        (extremeness (cdr b)))))))
      (take* wlist num))))

;; Calculate probability of the file being spam.  Returns three values:
;; the probability, the language of the file, and the list of 15
;; significant words with each "spamness".
(define (spamness-of-mail file)
  (receive (tokens lang) (tokenize-email file)
    (let1 words (most-interesting-words lang tokens 15)
      (values (spam-probability (map cdr words)) lang words))))

;; Predicate API
(define (mail-is-spam? file)
  (receive (prob lang words) (spamness-of-mail file)
    (>= prob 0.9)))

;;============================================================
;; Tokenizing message
;;

(define report warn)                    ;;for now

;; Entry point.  Reads FILE and returns two values; the list of
;; tokens and the language of the message.
;; First it tries to honor the charset in the message body
;; (if no charset is given, try to use "*jp" guessing), and if 
;; it fails, re-try with us-ascii.
;; (Because of retry, we need to accumulate tokens instead of
;; processing them as we read FILE.)
(define (tokenize-email file)

  (define (try charset)
    (let* ((tokens '())
           (lang (with-input-from-file file
                   (cut with-port-locking (current-input-port)
                        (cut tokenize-message file charset
                             (lambda (token) (push! tokens token))))))
           )
      (values (reverse! tokens) lang)))

  (with-error-handler
      (lambda (e)
        (report "error during processing ~a (~a)" file (ref e 'message))
        (values '() #f))
    (lambda ()
      (with-error-handler
          (lambda (e) (try "us-ascii"))
        (lambda () (try #f)))))
  )

;; Tokenize one chunk of message (header + body).  If the body is
;; MIME, call tokenize-mime, which in turn calls tokenise-message
;; recursively to deal with each part.  Returns a language tag
;; (for now, either #t or 'jp).
(define (tokenize-message file charset receiver)
  (let* ((headers (rfc822-header->list (current-input-port)))
         (content-type (get-content-type headers))
         (boundary (get-mime-boundary content-type))
         (charset (or charset (get-charset content-type)))
         )
    (tokenize-header headers receiver)
    (if boundary
        (tokenize-mime file receiver charset boundary)
        (tokenize-body file receiver charset content-type
                       (get-transfer-encoding headers)))
    ))

;; Tokenize a message header.  Returns void.
(define (tokenize-header headers receiver)
  (for-each (lambda (header)
              (with-input-from-string
                  (decode-field (string-join (map x->string (cdr header)) " ")
                                (x->string (gauche-character-encoding)))
                (cut tokenize-port receiver)))
            headers))

;; Tokenize a message body.
;; Returns a language tag, or #f.
(define (tokenize-body file receiver charset content-type transfer-encoding)
  (define (do-tokenize)
    (cond
     ;; Lots of emails that claims charset=us-ascii actually contains
     ;; iso-8859 characters.
     ((or (string-ci=? "us-ascii" charset)
          (string-prefix-ci? "iso-8859" charset))
      (tokenize-port-iso8859 receiver))
     ;; Deal with multibyte messages.
     ((or (equal? charset "*jp")
          (ces-conversion-supported? charset
                                     (x->string (gauche-character-encoding))))
      (with-input-from-port (open-input-conversion-port
                             (current-input-port) charset)
        (cut tokenize-port receiver)))
     (else
      (report "~a: unknown charset (~a): applying single-byte analysis"
              file charset)
      (tokenize-port-iso8859 receiver))
     ))
  
  (cond
   ((and content-type
         (not (or (string-prefix-ci? "text/" content-type)
                  (string-prefix-ci? "message/rfc822" content-type))))
    (report "skipping ~a (content-type=~a)" file content-type)
    #f)
   ((and transfer-encoding (string-ci=? transfer-encoding "base64"))
    (with-input-from-string (base64-decode-string (port->string (current-input-port)))
      do-tokenize))
   ((and transfer-encoding (string-ci=? transfer-encoding "quoted-printable"))
    (with-input-from-string (quoted-printable-decode-string (port->string (current-input-port)))
      do-tokenize))
   (else (do-tokenize))
   ))

;; Tokenize MIME message.  Calls tokenize-message for each part.
;; Returns a language tag, or #f.
(define (tokenize-mime file receiver charset boundary)
  ;; skip the toplevel part
  (define (skip-preamble line)
    (cond ((eof-object? line) #f)
          ((equal? line boundary) #t)
          (else (skip-preamble (read-line)))))
  ;; collect parts
  (define (collect-parts parts)
    (let loop ((parts '()))
      (let loop2 ((line (read-byte-line))
                  (lines '()))
        (cond ((eof-object? line)
               (cons (string-join (reverse! lines) "\r\n") parts))
              ((equal? (string-complete->incomplete line)
                       (string-complete->incomplete boundary))
               (loop (cons (string-join (reverse! lines) "\r\n") parts)))
              (else
               (loop2 (read-byte-line) (cons line lines)))))))
  ;; language
  (define lang #t)
  ;; main body
  (and (skip-preamble (read-line))
       (fold (lambda (part number)
               (with-input-from-string part
                 (lambda ()
                   (let1 lg
                       (tokenize-message #`",|file|#,|number|"
                                         charset receiver)
                     (when (symbol? lg) (set! lang lg)))))
               (+ number 1))
             0
             (reverse! (collect-parts '())))
       lang)
  )

;; Tokenizer utilities

;; Taken from scmail by Satoru Takabayashi
(define (decode-field str to-code)
  (with-error-handler 
   (lambda (e) str)
   (lambda ()
     (regexp-replace-all #/=\?([^?]+)\?([BQ])\?([^?]+)\?=\s*/ 
			 str
			 (lambda (m)
			   (let* ((charcode (m 1))
				  (encoding (m 2))
				  (message  (m 3))
				  (decode (if (equal? encoding "B")
					      base64-decode-string
					      quoted-printable-decode-string)))
			     (ces-convert (decode message)
					  charcode
					  to-code)))))))


(define (flush-token lis receiver)
  (unless (or (null? lis) (null? (cdr lis))
              (every (cut char-set-contains? #[\d_-] <>) lis))
    (receiver (list->string (map char-downcase (reverse! lis))))))

;; Base tokenizer for Japanese message.
(define (tokenize-port receiver)
  (define (alnum? ch) (char-set-contains? #[-$'!\w] ch))
  (define (mb? ch) (>= (char->integer ch) 256))
  (define (ascii ch acc)
    (cond ((eof-object? ch) (flush-token acc receiver))
          ((mb? ch)
           (flush-token acc receiver)
           (multibyte (read-char) ch))
          ((alnum? ch) (ascii (read-char) (cons ch acc)))
          (else 
           (flush-token acc receiver)
           (blank (read-char)))))
  (define (blank ch)
    (cond ((eof-object? ch))
          ((mb? ch) (multibyte (read-char) ch))
          ((alnum? ch) (ascii (read-char) (list ch)))
          (else (blank (read-char)))))
  (define (kanji? ch)
    (char-set-contains? #[-] ch))
  (define (multibyte ch prev)
    (cond ((eof-object? ch))
          ((or (eqv? ch #\return) (eqv? ch #\newline))
           (multibyte (read-char) prev)) ;; ignore newlines between mb char
          ((mb? ch)
           (cond ((char-set-contains? #[] ch)
                  (multibyte (read-char) #f))
                 (prev
                  (unless (and (not (kanji? prev))
                               (kanji? ch))
                    (flush-token (list ch prev) receiver))
                  (multibyte (read-char) ch))
                 (else
                  (multibyte (read-char) ch))))
          ((alnum? ch) (ascii (read-char) (list ch)))
          (else (blank (read-char)))))
  (ascii (read-char) '())
  'jp)

;; Base tokenizer for ASCII message.
(define (tokenize-port-iso8859 receiver)
  (define (alnum? b) (char-set-contains? #[-$'!\w] (integer->char b)))
  (define (ascii b acc)
    (cond ((eof-object? b) (flush-token acc receiver))
          ((alnum? b) (ascii (read-byte) (cons (integer->char b) acc)))
          (else (flush-token acc receiver)
                (blank (read-byte)))))
  (define (blank b)
    (cond ((eof-object? b))
          ((alnum? b) (ascii (read-byte) (list (integer->char b))))
          (else (blank (read-byte)))))
  (ascii (read-byte) '())
  #t)

;; helper function to get a line as a byte string
(define (read-byte-line . maybe-port)
  (let ((port (get-optional maybe-port (current-input-port)))
        (out  (open-output-string)))
    (let loop ((byte (read-byte port)))
      (cond ((eof-object? byte) byte)
            ((= byte #x0a) (get-output-string out))
            ((= byte #x0d) (loop (read-byte port)))
            (else (write-byte byte out) (loop (read-byte port)))))))

(define (get-content-type headers)
  (cond ((assoc "content-type" headers) => cadr) (else #f)))

(define (get-transfer-encoding headers)
  (cond ((assoc "content-transfer-encoding" headers) => cadr) (else #f)))

(define (get-mime-boundary content-type)
  (cond ((and content-type
              (#/multipart\/\w+\;.*boundary=\s*\"?([-=_\w().: ]+)\"?/i
               content-type))
         => (lambda (m) (string-append "--" (m 1))))
        (else #f)))

(define (get-charset content-type)
  (cond ((and content-type
              (#/charset=\"?([-_\w.]+)\"?/ content-type))
         => (cut <> 1))
        (else "*jp"))) ;;guess


;;============================================================
;; For testing
;;

(define (simple-stat)
  (let* ((tab (token-table))
         (mcount (ref tab 'message-count))
         (tcount (ref tab 'total-token-count))
         (totals (make-list *num-values* 0)))
    (format #t "lang       nonspam           spam\n")
    (dolist (lang *languages*)
      (let1 v (list (ref tcount (nonspam-table lang))
                    (ref mcount (nonspam-table lang))
                    (ref tcount (spam-table lang))
                    (ref mcount (spam-table lang)))
        (apply format #t "~4a: ~7dw/~5dm  ~7dw/~5dm\n" lang v)
        (set! totals (map + totals v))))
    (apply format #t "total:~7dw/~5dm  ~7dw/~5dm\n" totals)))

(define (test-spamness-of-mail email)
  (receive (prob lang words) (spamness-of-mail email)
    (print email)
    (print #`"  ,prob")
    (for-each (lambda (w) (print #`"    ,(car w) : ,(cdr w)"))
              words)))

;; (test-spamness-in-folder "~/Mail/inbox" #f)
;; (test-spamness-in-folder "~/Mail/spam" #t)
(define (test-spamness-in-folder directory expect-spam?)
  (let ((threshold 0.9)
        (count 0)
        (bad  '()))
    (for-each (lambda (email)
                (inc! count)
                (receive (prob lang words) (spamness-of-mail email)
                  (when (or (and expect-spam?
                                 (< prob threshold))
                            (and (not expect-spam?)
                                 (>= prob threshold)))
                    (push! bad (list email prob words)))))
              (directory-list (expand-path directory)
                              :children? #t :add-path? #t
                              :filter #/^\d+$/))
    (print #`"Out of ,count messages in ,directory")
    (print #`"  ,(length bad) messages are identified incorrectly:")
    (for-each (lambda (entry)
                (print #`"  ,(sys-basename (car entry)) (score=,(cadr entry)):")
                (for-each (lambda (w)
                            (print #`"    ,(car w) : ,(cdr w)"))
                          (caddr entry)))
              bad)
    ))


(provide "bayesian-filter")
